/*
 * Copyright 1991 Michael S. Warren and John K. Salmon.  All Rights Reserved.
 */

#include <stdio.h>		/* only use sprintf */
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stddef.h>
#include <unistd.h>
#include <errno.h>
#include "fastflpt.h"
#include "Assert.h"
#include "SDF.h"
#include "protos.h"
#include "macr.h"
#include "getparam.h"
#include "malloc.h"
#include "Malloc.h"
#include "SDFwrite.h"
#include "SDFread.h"
#include "physics_n.h"
#include "vop.h"
#include "Msgs.h"
#include "timers.h"
#include "singlio.h"
#include "mpmy.h"
#include "mpmy_io.h"
#include "mpmy_abnormal.h"
#include "gc.h"
#include "files.h"
#include "verify.h"
#include "memfile.h"

struct cosmo_s{
  float t;
  float a;
  float H0;
  float Omega0;
  float Lambda;
  float GNewt;
  float Zel_f;		/* the 'f' factor for linearly growing modes,
			   used only in set_vel = 1/H*Ddot/D.  It's
			   very close to 1 (exactly?) for flat models. */
} cosmo;

static void CosmoPush(struct cosmo_s *p, float time);
static SDF *startup(int argc, char **argv);
static int maxmem(void);
static int maxheap(void);
static void SanityCheck(body *btab, int nobj, int gnobj, double *mtotp);

void
main(int argc, char *argv[])
{
  int gnobj, nobj;
  body *btab;
  int npixels;
  char outnamebase[256];
  SDF *csdfp;			/* SDF pointer to control file */
  SDF *sdfp;
  int massconf, xconf, yconf, zconf;
  int vxconf, vyconf, vzconf;
  int dmconf;
  char name[256];
  int read_nfiles;
  int do_periodic;
  int timeout;
  double mtot;
  int sum;

  MPMY_Init(&argc, &argv);
  atexit(MPMY_Abchdir);
  singlPrintf("Welcome to the N-body graphics interface running on %d procs\n",
	      MPMY_Nproc());
  csdfp = startup(argc, argv);
  SDFgetintOrDefault(csdfp, "timeout", &timeout, 600);
  if (timeout > 0) MPMY_TimeoutSet(timeout);
  SDFgetstring(csdfp, "datafile", name, sizeof(name));
  SDFgetintOrDefault(csdfp, "do_periodic", &do_periodic, 0);
  SDFgetintOrDefault(csdfp, "read_nfiles", &read_nfiles, 0);
  /* SDFsetbufsz(65536); */
  if( !((strncmp(name, "test", 4) == 0)) ) {
    singlPrintf("Reading \"%s\"\n", name);
    if (read_nfiles) MPMY_Nfileio(1);
    EnableTimer(&SDFreadTm, "SDFread");
    sdfp = SDFread(csdfp, (void **)&btab, &gnobj, &nobj, sizeof(body),
		   "mass", offsetof(body, mass), &massconf,
		   "x", offsetof(body, pos[0]), &xconf,
		   "y", offsetof(body, pos[1]), &yconf,
		   "z", offsetof(body, pos[2]), &zconf,
		   "vx", offsetof(body, vel[0]), &vxconf,
		   "vy", offsetof(body, vel[1]), &vyconf,
		   "vz", offsetof(body, vel[2]), &vzconf,
		   NULL);
    dmconf = 0;
    OutputTimer(&SDFreadTm, singlPrintf); /* global sync and sets timer->max */
    singlPrintf("read speed %.0f kb/s\n", gnobj*8*sizeof(float)/(1000.0*SDFreadTm.max));
    DisableTimer(&SDFreadTm);
    sum = cksum(btab, nobj*sizeof(body));
    MPMY_Combine(&sum, &sum, 1, MPMY_INT, MPMY_BXOR);
    singlPrintf("Checksum of body data: 0x%x\n", sum);
    if (read_nfiles) MPMY_Nfileio(0);
    Msgf(("Data read, nobj=%d, gnobj=%d\n", nobj, gnobj));
    Msgf(("Nproc:%d, Procnum: %d\n",
	  MPMY_Nproc(), MPMY_Procnum()));
    if( dmconf && massconf ){
      SinglError("You can't have both 'dm' and 'mass' in the input file\n");
    }
    if( dmconf==0 && massconf==0 ){
      SinglError("You must have one of 'dm' or 'mass' in the input file\n");
    }

    if( xconf==0 || yconf==0 || zconf==0 ){
      SinglError("Could not find %s %s %s in data file!\n",
		 (xconf==0)? "x" : "",
		 (yconf==0)? "y" : "",
		 (zconf==0)? "z" : "");
    }
    if( vxconf != vyconf || vxconf != vzconf ){
      SinglError("Found only some velocity components!\n");
    }

    if(sdfp)
      SDFclose(sdfp);
  }

  singlPrintf("Maxmem after data read is %d (%d)\n", maxmem(), maxheap());
  if( Msg_test("memleak") ){
    Msg_do("Memory map after data read\n");
    malloc_print();
  }

  SDFgetstringOrDie(csdfp, "outfile", outnamebase, sizeof(outnamebase));
  SDFgetintOrDefault(csdfp, "npixels", &npixels, 512);
  singlPrintf("Creating a map width=%d, height=%d\n", npixels, (npixels+1)/2);

  if(csdfp) 
    SDFclose(csdfp);

  singlFflush();
  SanityCheck(btab, nobj, gnobj, &mtot);

  eaimage(npixels, nobj, btab, outnamebase);
  Msgf(("Bye!\n"));
  Msg_flush();
  exit(0);			/* trex seems to hang in __exit() */
}

static SDF *startup(int argc, char **argv){
  SDF *csdfp;
  char msg_turn_on[512];
  char msgdir[256];
  char tmp[256];
  char *msgbase, *lastslash;
  char cfile[256];
  int Msg_memfile;

  if (argc > 1)
    strncpy(cfile, argv[1], sizeof(cfile));
  else
    Getsparam("control file", cfile);
  if ((csdfp = SDFopen(NULL, cfile)) == NULL) {
    SinglError("Sorry, couldn't SDFopen %s\n%s\n",
	       cfile, SDFerrstring);
  }
  singlPrintf("cfile \"%s\" opened\n", cfile);
  SDFgetintOrDefault(csdfp, "Msg_memfile", &Msg_memfile, 0);
  if (Msg_memfile) {
#if defined(__PARAGON__) || defined(_AIX) || defined(sun4)
    sigio_setup();
#endif
    singlPrintf("Putting all messages in memfile\n");
    memfile_init(Msg_memfile);
    Msg_addfile(0, (Msgvfprintf_t)memfile_vfprintf, 0);
    singlPrintf("done\n");
  } else {
    /* Get the msgdir either from:
       argv[2] 
       "msgbase" in csdfp
       misc.argv[0]/msg

       We then append .<procnum> to the name
       */
    if( argc > 2 ){
      msgbase = argv[2];
    }else if( SDFgetstring(csdfp, "msgbase", tmp, sizeof(tmp))==0 ){
      msgbase = tmp;
    }else{
      lastslash = strrchr(argv[0], '/');
      if( lastslash ){
	msgbase = lastslash+1;
      }else{
	msgbase = argv[0];
      }
      sprintf(tmp, "misc.%s/msg", msgbase);
      msgbase = tmp;
    }	
    sprintf(msgdir, "%s.%d", msgbase, MPMY_Procnum());
    singlPrintf("MsgdirInit(%s)\n", msgdir);
    MsgdirInit(msgdir);
    Msg_flushalways(1);
  }
  SDFgetstringOrDefault(csdfp, "Msg_turn_on", msg_turn_on, 
			sizeof(msg_turn_on), "");
  Msg_turnon(msg_turn_on);

  if( Msg_test("Malloc.c") ){
    malloc_debug(2);
    Msg_do("Malloc_debug(2), expect slow mallocs\n");
  }else{
    malloc_debug(1);
  }
  if( _Msg_enabled ){
    Msg_do("startup on physnode: %s\n", MPMY_Physnode());
  }

  singlPrintf("Enabling timers...");
  if( MPMY_Nproc() > 1 ){
    EnableCounter(&MPMYSendCnt, "MPMY Sends");
    EnableCounter(&MPMYRecvCnt, "MPMY Recvs");
    EnableCounter(&MPMYDoneCnt, "MPMY Done");
  }
  singlPrintf("done\n");
  return csdfp;
}

static void SanityCheck(body * btab, int nobj, int gnobj, double *mtotp){
  double mtot;
  body * p;
  int sumnobj;
  MPMY_Comm_request req;

  mtot = 0.0;
  for (p = btab; p < btab+nobj; p++) {
    mtot += p->mass;
    /* We really could use more checks here! */
  }
  sumnobj = nobj;
  MPMY_ICombine_Init(&req);
  MPMY_ICombine(&mtot, &mtot, 1, MPMY_DOUBLE, MPMY_SUM, req);
  MPMY_ICombine(&sumnobj, &sumnobj, 1, MPMY_INT, MPMY_SUM, req);
  MPMY_ICombine_Wait(req);
  assert(sumnobj == gnobj);
  Msgf(("Particle 0, %g, %g %g %g, %g %g %g\n",
	btab->mass, 
	btab->pos[0], btab->pos[1], btab->pos[2],
	btab->vel[0], btab->vel[1], btab->vel[2]));
  Msgf(("Particle %d, %g, %g %g %g, %g %g %g\n", nobj-1,
	btab[nobj-1].mass, 
	btab[nobj-1].pos[0], btab[nobj-1].pos[1], btab[nobj-1].pos[2],
	btab[nobj-1].vel[0], btab[nobj-1].vel[1], btab[nobj-1].vel[2]));
  singlPrintf("gnobj = %d, mtot = %f\n", gnobj, mtot);
  *mtotp = mtot;
}


#define one_kpc (3.08567802e16) /* km */
#define one_Gyr (3.1558149984e16) /* sec */

float 
Anow(float time)
{
  struct cosmo_s foo;

  foo = cosmo;
  CosmoPush(&foo, time);
  return foo.a;
}    

float
Znow(float time)
{
  return 1.F/Anow(time) - 1.F;
}

float
Hnow(float t)
{
  struct cosmo_s foo;
  float Omega0 = cosmo.Omega0;
  float Lambda = cosmo.Lambda;
  float H0 = cosmo.H0;
  float a2, a3;

  foo = cosmo;
  CosmoPush(&foo, t);
    
  a2 = foo.a*foo.a;
  a3 = foo.a*a2;
  singlPrintf("Hnow(t=%g), a=%g, H0=%g\n", t, foo.a, H0);
  return H0 * sqrt( Omega0/a3 + Lambda - (Omega0 + Lambda - 1.F)/a2);
}


static int 
maxheap(void)
{
  int memused = malloc_heapsz()/1024;
  MPMY_Combine(&memused, &memused, 1, MPMY_INT, MPMY_MAX);
  return memused;
}

static int 
maxmem(void)
{
  int memused = malloc_used()/1024;
  MPMY_Combine(&memused, &memused, 1, MPMY_INT, MPMY_MAX);
  return memused;
}

static void 
CosmoPush(struct cosmo_s *p, float time){
  float Omega0 = p->Omega0;
  float Lambda = p->Lambda;
  float H0 = p->H0;
  float H, a2, a3, aold, a2dot;
  int i;
  float deltat, dt;
  int nstep;

  /* The cosmo structure holds,H0, Omega0, Lambda' = Lambda/3H0^2, a
     and t.  We integrate (forward or backward) to the new 'time' */

  deltat = time - p->t;
  if( deltat == 0.F )
    return;

  /* Felten et al do all their integrals with dt=1/(400 H0).  We can
     do the same by choosing Nstep appropriately.  In fact, we can
     do a little better by ensuring dt < 1/(400 H). */
  aold = p->a;
  a2 = aold*aold;
  a3 = a2*aold;
  H = H0 * sqrt( Omega0/a3 + Lambda - (Omega0 + Lambda - 1.F)/a2);
  nstep = (int)(400.*H*fabs(deltat)) + 1;
  Msgf(("Cosmo push %d steps, deltat=%g, H*deltat=%g\n", 
	nstep, deltat, deltat*H));
  dt = deltat/(float)nstep;
    
  for(i=0; i<nstep; i++){
    aold = p->a;
    a2 = aold*aold;
    a3 = a2*aold;
    H = H0 * sqrt( Omega0/a3 + Lambda - (Omega0 + Lambda - 1.F)/a2);
    /* Follow the advice of Felten et al.  Do this to second-order */
#if 1
    a2dot = aold*H0*H0*( -0.5F*Omega0/a3 + Lambda);
#else
    a2dot = 0.F;
#endif
    p->a = aold + dt*H*aold + 0.5F*dt*dt*a2dot;
  }
  Msgf(("After push Z=%g\n", 1./p->a - 1.));
  p->t = time;
}

